from __future__ import print_function, unicode_literals

import os
import pathlib
from dataclasses import dataclass
from os.path import expanduser as user
from typing import Optional

import cowsay
from pyfiglet import Figlet
from PyInquirer import Separator
from PyInquirer.prompt import prompt
from yaspin import yaspin

from now.deployment.deployment import cmd
from now.system_information import get_system_state

cur_dir = pathlib.Path(__file__).parent.resolve()
NEW_CLUSTER = '🐣 create new'
AVAILABLE_SOON = 'will be available in upcoming versions'
QUALITY_MAP = {
    'medium': ('ViT-B32', 'openai/clip-vit-base-patch32'),
    'good': ('ViT-B16', 'openai/clip-vit-base-patch16'),
    'excellent': ('ViT-L14', 'openai/clip-vit-large-patch14'),
}

LIST_OF_DB = {
    'artworks',
    'birds',
    'cas',
    'chest x-ray',
    'deepfashion',
    'geolocation',
}


@dataclass
class UserInput:
    # data related
    dataset: Optional[str] = 'deepfashion'
    is_custom_dataset: Optional[bool] = False
    custom_dataset_type: Optional[str] = None
    dataset_secret: Optional[str] = None
    dataset_url: Optional[str] = None
    dataset_path: Optional[str] = None
    # model related
    model_quality: str = 'medium'
    model_variant: str = 'ViT-B32'
    # cluster related
    cluster: Optional[str] = None
    create_new_cluster: bool = False
    new_cluster_type: str = 'local'


def headline():
    f = Figlet(font='slant')
    print('Welcome to:')
    print(f.renderText('Jina NOW'))
    print('Get your search case up and running - end to end.')
    print(
        'We take your images and texts, train a model, push it to the jina hub, '
        'deploy a flow and a frontend in the cloud or locally.'
    )
    print(
        'If you want learn more about our framework please visit: https://docs.jina.ai/'
    )
    print(
        '💡 Make sure you assign give enough memory to your Docker daemon. '
        '5GB - 8GB should be okay.'
    )
    print()


def get_user_input(contexts, active_context, os_type, arch, **kwargs) -> UserInput:
    headline()
    user_input = UserInput()
    if kwargs and kwargs['data']:
        assign_data_fields(user_input, kwargs['data'])
    else:
        ask_data(user_input, **kwargs)
    ask_quality(user_input, **kwargs)
    ask_deployment(user_input, contexts, active_context, os_type, arch, **kwargs)
    return user_input


def prompt_plus(questions, attribute, **kwargs):
    if kwargs and kwargs[attribute]:
        return kwargs[attribute]
    else:
        answer = prompt(questions)
        if attribute in answer:
            return answer[attribute]
        else:
            os.system('clear')
            cowsay.cow('see you soon 👋')
            exit(0)


def assign_data_fields(user_input, data):
    user_input.dataset = 'custom'
    user_input.is_custom_dataset = True
    try:
        data = os.path.expanduser(data)
    except Exception:
        pass
    if os.path.exists(data):
        user_input.custom_dataset_type = 'path'
        user_input.dataset_path = data
    elif 'http' in data:
        user_input.custom_dataset_type = 'url'
        user_input.dataset_url = data
    elif data in LIST_OF_DB:
        user_input.dataset = data
        user_input.is_custom_dataset = False
    else:
        user_input.custom_dataset_type = 'docarray'
        user_input.dataset_secret = data


def ask_data(user_input: UserInput, **kwargs):
    questions = [
        {
            'type': 'list',
            'name': 'dataset',
            'message': 'What dataset do you want to use?',
            'choices': [
                {'name': '🖼 artworks (≈8K docs)', 'value': 'best-artworks'},
                {
                    'name': '💰 nft - bored apes (10K docs)',
                    'value': 'nft-monkey',
                },
                {'name': '🦆 birds (≈12K docs)', 'value': 'bird-species'},
                {'name': '🚗 cars (≈16K docs)', 'value': 'stanford-cars'},
                {
                    'name': '🏞 geolocation (≈50K docs)',
                    'value': 'geolocation-geoguessr',
                },
                {'name': '👕 fashion (≈53K docs)', 'value': 'deepfashion'},
                {
                    'name': '☢️ chest x-ray (≈100K docs)',
                    'value': 'nih-chest-xrays',
                },
                Separator(),
                {
                    'name': '✨ custom',
                    'value': 'custom',
                },
            ],
        },
    ]
    user_input.dataset = prompt_plus(questions, 'dataset')

    if user_input.dataset == 'custom':
        user_input.is_custom_dataset = True
        ask_data_custom(user_input, **kwargs)
    else:
        user_input.is_custom_dataset = False


def ask_data_custom(user_input: UserInput, **kwargs):
    questions = [
        {
            'type': 'list',
            'name': 'custom_dataset_type',
            'message': (
                'How do you want to provide input? (format: https://docarray.jina.ai/)'
            ),
            'choices': [
                {
                    'name': 'docarray.pull(...) (recommended)',
                    'value': 'docarray',
                },
                {
                    'name': 'docarray URL',
                    'value': 'url',
                },
                {
                    'name': 'local mounted path',
                    'value': 'path',
                    'disabled': AVAILABLE_SOON,
                },
            ],
        },
    ]
    custom_dataset_type = prompt_plus(questions, 'custom_dataset_type')
    user_input.custom_dataset_type = custom_dataset_type

    if custom_dataset_type == 'docarray':
        questions = [
            {
                'type': 'password',
                'name': 'secret',
                'message': 'Please enter your docarray secret',
            },
        ]
        user_input.dataset_secret = prompt_plus(questions, 'secret')
    elif custom_dataset_type == 'url':
        questions = [
            {
                'type': 'input',
                'name': 'url',
                'message': 'Please paste in your URL for the docarray:',
            },
        ]
        user_input.dataset_url = prompt_plus(questions, 'url')
    else:
        pass


def ask_quality(user_input: UserInput, **kwargs):
    questions = [
        {
            'type': 'list',
            'name': 'quality',
            'message': 'What quality do you expect?',
            'choices': [
                {'name': '🦊 medium (≈3GB mem, 15q/s)', 'value': 'medium'},
                {'name': '🐻 good (≈3GB mem, 2.5q/s)', 'value': 'good'},
                {
                    'name': '🦄 excellent (≈4GB mem, 0.5q/s)',
                    'value': 'excellent',
                },
            ],
            'filter': lambda val: val.lower(),
        }
    ]
    quality = prompt_plus(questions, 'quality', **kwargs)

    if quality == 'medium':
        print('  🚀 you trade-off a bit of quality for having the best speed')
    elif quality == 'good':
        print('  ⚖️ you have the best out of speed and quality')
    elif quality == 'excellent':
        print('  ✨ you trade-off speed to having the best quality')

    user_input.model_quality, user_input.model_variant = QUALITY_MAP[quality]


def get_context_names(contexts, active_context=None):
    names = [c['name'] for c in contexts] if contexts is not None else []
    if active_context is not None:
        names.remove(active_context['name'])
        names = [active_context['name']] + names
    return names


def ask_deployment(
    user_input: UserInput, contexts, active_context, os_type, arch, **kwargs
):
    choices = (get_context_names(contexts, active_context)) + [NEW_CLUSTER]

    questions = [
        {
            'type': 'list',
            'name': 'cluster',
            'message': 'Where do you want to deploy your search engine?',
            'choices': choices,
        }
    ]
    cluster = prompt_plus(questions, 'cluster', **kwargs)
    user_input.cluster = cluster

    if cluster == NEW_CLUSTER:
        user_input.cluster = None
        user_input.create_new_cluster = True
        questions = [
            {
                'type': 'list',
                'name': 'cluster_new',
                'message': 'Where do you want to create a new cluster?',
                'choices': [
                    {
                        'name': '📍 local (Kubernetes in Docker)',
                        'value': 'local',
                    },
                    {'name': '⛅️ Google Kubernetes Engine', 'value': 'gke'},
                    {
                        'name': '⛅️ Amazon Elastic Kubernetes Service',
                        'disabled': AVAILABLE_SOON,
                    },
                    {
                        'name': '⛅️ Azure Kubernetes Service',
                        'disabled': AVAILABLE_SOON,
                    },
                    {
                        'name': '⛅️ DigitalOcean Kubernetes',
                        'disabled': AVAILABLE_SOON,
                    },
                ],
                'filter': lambda val: val.lower(),
            }
        ]
        user_input.new_cluster_type = prompt_plus(questions, 'cluster_new')
        if user_input.new_cluster_type == 'gke':
            out, _ = cmd('which gcloud')
            if not out:
                if not os.path.exists(user('~/.cache/jina-now/google-cloud-sdk')):
                    with yaspin(text='Setting up gcloud', color='green') as spinner:
                        cmd(
                            f'/bin/bash {cur_dir}/scripts/install_gcloud.sh {os_type} {arch}',
                        )
                        spinner.ok('🛠️')


if __name__ == '__main__':
    print(get_user_input(*get_system_state()))



import io
import math
from typing import Callable

import clip
import torch
from docarray import Document
from jina import DocumentArray
from PIL import Image
from tqdm import tqdm

from now.data_loading.utils import upload_to_gcloud_bucket


def _text_collate_fn(batch: DocumentArray):
    return clip.tokenize([doc.text.lower() for doc in batch], truncate=True)


class _ImageCollateFn:
    def __init__(self, now_preprocess: Callable):
        self._now_preprocess = now_preprocess

    def __call__(self, batch: DocumentArray):
        return torch.stack(
            [
                self._now_preprocess(
                    Image.fromarray(doc.tensor)
                    if doc.tensor
                    else Image.open(io.BytesIO(doc.blob))
                )
                for doc in batch
            ],
            dim=0,
        )


def embed_docs(
    model: Callable,
    docs: DocumentArray,
    collate_fn: Callable,
    device: str = 'cpu',
    batch_size: int = 128,
):
    total = math.ceil(len(docs) / batch_size)

    with torch.no_grad():
        for batch in tqdm(docs.batch(batch_size=batch_size), total=total):
            x = collate_fn(batch)
            x = x.to(device)
            y = model(x)
            batch.embeddings = y.cpu().detach().numpy()


def to_jpg(image_docs):
    def pil2bytes(img: Image):
        img_byte_arr = io.BytesIO()
        img.save(img_byte_arr, format="JPEG", quality=75)
        return img_byte_arr.getvalue()

    def convert_to_jpeg(d: Document):
        if not d.blob:
            d.convert_image_tensor_to_blob()
            im = Image.fromarray(d.tensor)
            d.tensor = None
            d.blob = pil2bytes(im)
        return d

    image_docs.apply(convert_to_jpeg)


def embed_dataset(
    dataset: str,
    model: str,
    project: str,
    bucket: str,
    location: str,
    batch_size: int = 128,
    num_workers: int = 8,
):
    path = f'{dataset}.bin'

    print(f'===> {dataset} - {model}')
    print(f'  Loading {dataset} dataset from {path} ...')
    docs = DocumentArray.load_binary(path)
    print(f'  Dataset size: {len(docs)}')

    print(f'  Loading {model} CLIP model ...')
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    clip_model, preprocess = clip.load(model, device)
    clip_model.eval()
    clip_model.to(device)

    print(f'  Splitting text and image docs ...')
    text_docs = DocumentArray(
        [doc for doc in docs if doc.tags['content_type'] == 'text']
    )
    image_docs = DocumentArray(
        [doc for doc in docs if doc.tags['content_type'] == 'image']
    )
    print(f'  Num text docs: {len(text_docs)}')
    print(f'  Num image docs: {len(image_docs)}')

    print(f'  Embedding text docs ...')
    embed_docs(
        clip_model.encode_text,
        text_docs,
        collate_fn=_text_collate_fn,
        device=device,
        batch_size=batch_size,
    )
    print(f'  Done!')

    print(f'  Embedding image docs ...')
    embed_docs(
        clip_model.encode_image,
        image_docs,
        collate_fn=_ImageCollateFn(preprocess),
        device=device,
        batch_size=batch_size,
    )
    print(f'  Done!')

    print(f' convert images to jpeg for smaller datasets')
    to_jpg(image_docs)

    print(f'  Saving embedded docs ...')
    docs = text_docs + image_docs
    docs = docs.shuffle(42)
    out = f'{dataset}.{model.replace("/", "")}.bin'
    docs.save_binary(out)
    print(f'  Saved embedded docs to {out} ...')

    print(f'  Uploading dataset ...')
    upload_to_gcloud_bucket(project, bucket, location, out)
    print(f'  Uploaded dataset to gs://{bucket}/{location}/{out}')


def main():
    """
    Main method.
    """
    project = 'jina-simpsons-florian'
    bucket = 'jina-fashion-data'
    location = 'data/one-line/datasets'
    datasets = [
        'tll',
        'nft-monkey',
        'deepfashion',
        'nih-chest-xrays',
        'geolocation-geoguessr',
        'stanford-cars',
        'bird-species',
        'best-artworks',
    ]
    models = ['ViT-B/32', 'ViT-B/16', 'ViT-L/14']
    batch_size = 128

    for dataset in datasets:
        for model in models:
            embed_dataset(dataset, model, project, bucket, location, batch_size)


if __name__ == '__main__':
    main()



import tempfile

import cowsay
from yaspin import yaspin

from now import run_backend, run_frontend
from now.cloud_manager import setup_cluster
from now.deployment.deployment import cmd
from now.dialog import get_context_names, get_user_input, prompt_plus
from now.system_information import get_system_state


def stop_now(contexts, active_context, **kwargs):
    choices = get_context_names(contexts, active_context)
    if len(choices) == 0:
        cowsay.cow('nothing to stop')
        return
    elif len(choices) == 1:
        cluster = choices[0]
    else:
        questions = [
            {
                'type': 'list',
                'name': 'cluster',
                'message': 'Which cluster do you want to delete?',
                'choices': choices,
            }
        ]
        cluster = prompt_plus(questions, 'cluster')
    if cluster == 'kind-jina-now':
        with yaspin(text=f"Remove local cluster {cluster}", color="green") as spinner:
            cmd(f'{kwargs["kind_path"]} delete clusters jina-now')
            spinner.ok('💀')
        cowsay.cow('local jina NOW cluster removed')
    else:
        with yaspin(text=f"Remove jina NOW from {cluster}", color="green") as spinner:
            cmd(f'{kwargs["kubectl_path"]} delete ns nowapi')
            spinner.ok('💀')
        cowsay.cow(f'nowapi namespace removed from {cluster}')


def run_k8s(os_type='linux', arch='x86_64', **kwargs):
    contexts, active_context, is_debug = get_system_state()
    if kwargs['cli'] == 'stop':
        stop_now(contexts, active_context, **kwargs)
    else:
        with tempfile.TemporaryDirectory() as tmpdir:
            docker_frontend_tag = '0.0.1'

            user_input = get_user_input(
                contexts, active_context, os_type, arch, **kwargs
            )
            setup_cluster(user_input.cluster, user_input.new_cluster_type, **kwargs)
            (
                gateway_host,
                gateway_port,
                gateway_host_internal,
                gateway_port_internal,
            ) = run_backend.run(user_input, is_debug, tmpdir, **kwargs)
            frontend_host, frontend_port = run_frontend.run(
                user_input.dataset,
                gateway_host,
                gateway_port,
                gateway_host_internal,
                gateway_port_internal,
                docker_frontend_tag,
                tmpdir,
                **kwargs,
            )
            url = f'{frontend_host}' + (
                '' if str(frontend_port) == '80' else f':{frontend_port}'
            )
            print()

            # print(f'✅ Your search case running.\nhost: {node_ip}:30080')
            # print(f'host: {node_ip}:30080')
            cowsay.cow(f'You made it:\n{url}')


if __name__ == '__main__':
    run_k8s()



import sys
from os import path
import platform
from setuptools import find_packages
from setuptools import setup
import pkg_resources

if sys.version_info < (3, 7, 0):
    raise OSError(f'Jina NOW requires Python >=3.7, but yours is {sys.version}')

try:
    pkg_name = 'jina-now'
    libinfo_py = path.join('now', '__init__.py')
    libinfo_content = open(libinfo_py, 'r', encoding='utf8').readlines()
    version_line = [l.strip() for l in libinfo_content if l.startswith('__version__')][
        0
    ]
    exec(version_line)  # gives __version__
except FileNotFoundError:
    __version__ = '0.0.0'

try:
    with open('README.md', encoding='utf8') as fp:
        _long_description = fp.read()
except FileNotFoundError:
    _long_description = ''

import pathlib


with pathlib.Path('requirements_pip.txt').open() as requirements_txt:
    install_requires = [
        str(requirement)
        for requirement in pkg_resources.parse_requirements(requirements_txt)
    ]
sys_platform = platform.system().lower()
if sys_platform == 'darwin':
    torch_requirement = 'torch==1.10.2'
else:
    torch_requirement = 'torch @ https://download.pytorch.org/whl/cpu/torch-1.10.2%2Bcpu-cp37-cp37m-linux_x86_64.whl'

install_requires.append(torch_requirement)

setup(
    name=pkg_name,
    packages=find_packages(),
    version=__version__,
    include_package_data=True,
    description='Jina NOW - get your neural search case up and running in minutes.',
    author='Jina AI',
    author_email='hello@jina.ai',
    license='Apache 2.0',
    url='https://github.com/jina-ai/now/',
    download_url='https://github.com/jina-ai/now/tags',
    long_description=_long_description,
    long_description_content_type='text/markdown',
    zip_safe=False,
    setup_requires=['setuptools>=18.0', 'wheel'],
    install_requires=install_requires,
    entry_points={"console_scripts": ["jina-now = now.cli:cli"]},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Environment :: Console',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    project_urls={
        'Documentation': 'https://docs.jina.ai',
        'Source': 'https://github.com/jina-ai/now',
        'Tracker': 'https://github.com/jina-ai/now/issues',
    },
    keywords='jina neural-search neural-network deep-learning now private data democratization',
)



import csv
import json
import multiprocessing as mp
import os
import re
from dataclasses import dataclass, field
from random import shuffle
from typing import Any, Dict, Optional

from jina import Document, DocumentArray
from tqdm import tqdm

from now.data_loading.utils import upload_to_gcloud_bucket

IMAGE_SHAPE = (224, 224)


@dataclass
class _DataPoint:
    id: str
    text: Optional[str] = None
    image_path: Optional[str] = None
    content_type: str = 'image'
    label: str = ''
    split: str = 'none'
    tags: Dict[str, Any] = field(default_factory=lambda: {})


def _build_doc(datapoint: _DataPoint) -> Document:
    doc = Document(id=datapoint.id)
    if datapoint.content_type == 'image':
        doc.uri = datapoint.image_path
        doc.load_uri_to_image_tensor()
        doc.set_image_tensor_shape(IMAGE_SHAPE)
    else:
        doc.text = datapoint.text
    doc.tags = {'finetuner_label': datapoint.label, 'split': datapoint.split}
    doc.tags.update(datapoint.tags)
    doc.tags.update({'content_type': datapoint.content_type})
    return doc


def _build_deepfashion(root: str, num_workers: int = 8) -> DocumentArray:
    """
    Build the deepfashion dataset.
    Download the raw dataset from
    https://drive.google.com/drive/folders/0B7EVK8r0v71pVDZFQXRsMDZCX1E?resourcekey=0-4R4v6zl4CWhHTsUGOsTstw
    :param root: the dataset root folder.
    :param num_workers: the number of parallel workers to use.
    :return: DocumentArray
    """

    extension = '.jpg'
    imagedir = os.path.join(root, 'Img')
    fsplit = os.path.join(root, 'Eval', 'list_eval_partition.txt')
    fcolors = os.path.join(root, 'Anno', 'attributes', 'list_color_cloth.txt')

    # read list_eval_partition.txt
    img2split = {}
    with open(fsplit, 'r') as f:
        for line in f.read().splitlines()[2:]:
            img, _, split, _ = re.split(r' +', line)
            img2split[img] = split

    # read list_color_cloth.txt
    img2color = {}
    with open(fcolors, 'r') as f:
        for line in f.read().splitlines()[2:]:
            img, color, *_ = re.split(r'  +', line)
            img2color[img] = color

    # add image docs
    data = []
    for rootdir, _, fnames in os.walk(imagedir):
        labels = []
        productid = os.path.relpath(rootdir, imagedir)
        for fname in fnames:
            if fname.endswith(extension):
                path = os.path.join(rootdir, fname)
                imgid = os.path.relpath(path, imagedir)
                split = img2split[imgid]
                color = img2color[imgid]
                label = productid + '/' + color
                labels.append(label)
                data.append(
                    _DataPoint(
                        id=imgid,
                        image_path=path,
                        label=label,
                        split=split,
                        tags={'color': color},
                    )
                )

        # add text doc
        if len(labels) > 0:
            for label in set(labels):
                _, gender, category, _, color = label.split('/')
                text_elements = [category, gender, color]
                shuffle(text_elements)
                text = (
                    f'{" ".join(text_elements)}'.lower()
                    .replace('-', ' ')
                    .replace('_', ' ')
                )
                data.append(
                    _DataPoint(
                        id=rootdir,
                        text=text,
                        content_type='text',
                        label=label,
                        tags={'color': color},
                    )
                )

    # build docs
    with mp.Pool(processes=num_workers) as pool:
        docs = list(tqdm(pool.imap(_build_doc, data)))

    return DocumentArray(docs)


def _build_nih_chest_xrays(root: str, num_workers: int = 8) -> DocumentArray:
    """
    Build the NIH chest xrays dataset.
    Download the raw dataset from
    https://www.kaggle.com/nih-chest-xrays/data
    :param root: the dataset root folder.
    :param num_workers: the number of parallel workers to use.
    :return: DocumentArray
    """

    extension = '.png'
    flabels = 'Data_Entry_2017.csv'
    ftrain = 'train_val_list.txt'
    ftest = 'test_list.txt'

    # read Data_Entry_2017.csv
    # labels - fname: (finding, patient id)
    with open(os.path.join(root, flabels), 'r') as f:
        reader = csv.reader(f)
        next(reader)
        labels = {row[0]: (row[1], row[3]) for row in reader}

    # read train_val_list.txt
    with open(os.path.join(root, ftrain), 'r') as f:
        train_list = f.read().splitlines()

    # read test_list.txt
    with open(os.path.join(root, ftest), 'r') as f:
        test_list = f.read().splitlines()

    # add image docs
    data = []
    for rootdir, _, fnames in os.walk(root):
        for fname in fnames:
            if fname.endswith(extension):

                path = os.path.join(rootdir, fname)
                label = labels.get(fname)[0]  # or labels[1]
                if fname in train_list:
                    split = 'train'
                elif fname in test_list:
                    split = 'test'
                else:
                    raise ValueError(
                        f'Doc with fname: {fname} not in train or test splits'
                    )
                data.append(
                    _DataPoint(id=fname, image_path=path, label=label, split=split)
                )

    # add text docs
    labelnames = {label for _, (label, __) in labels.items()}
    for label in labelnames:
        data.append(
            _DataPoint(
                id=label,
                text=label.lower()
                .replace('|', ' ')
                .replace('_', ' ')
                .replace('-', ' '),
                content_type='text',
                label=label,
            )
        )

    # build docs
    with mp.Pool(processes=num_workers) as pool:
        docs = list(tqdm(pool.imap(_build_doc, data)))

    return DocumentArray(docs)


def _build_geolocation_geoguessr(root: str, num_workers: int = 8) -> DocumentArray:
    """
    Build the geolocation-geoguessr dataset.
    Download the raw dataset from
    https://www.kaggle.com/ubitquitin/geolocation-geoguessr-images-50k
    :param root: the dataset root folder.
    :param num_workers: the number of parallel workers to use.
    :return: DocumentArray
    """

    extension = '.jpg'

    # add image docs
    data = []
    for rootdir, _, fnames in os.walk(root):
        label = os.path.relpath(rootdir, root)
        for fname in fnames:
            if fname.endswith(extension):
                path = os.path.join(rootdir, fname)
                data.append(_DataPoint(id=fname, image_path=path, label=label))

        # add text doc
        if len(fnames) > 0:
            data.append(
                _DataPoint(
                    id=label, text=label.lower(), content_type='text', label=label
                )
            )

    # build docs
    with mp.Pool(processes=num_workers) as pool:
        docs = list(tqdm(pool.imap(_build_doc, data)))

    return DocumentArray(docs)


def _build_stanford_cars(root: str, num_workers: int = 8) -> DocumentArray:
    """
    Build the stanford cars dataset.
    Download the raw dataset from
    https://www.kaggle.com/jessicali9530/stanford-cars-dataset
    :param root: the dataset root folder.
    :param num_workers: the number of parallel workers to use.
    :return: DocumentArray
    """

    extension = '.jpg'
    train_data = os.path.join(root, 'car_data', 'train')
    test_data = os.path.join(root, 'car_data', 'test')

    # add image docs
    data = []
    labels = []
    for split, root in [('train', train_data), ('test', test_data)]:
        for rootdir, _, fnames in os.walk(root):
            if len(fnames) > 0:
                label = os.path.relpath(rootdir, root)
                labels.append(label)
                for fname in fnames:
                    if fname.endswith(extension) and 'cropped' not in fname:
                        path = os.path.join(rootdir, fname)
                        data.append(
                            _DataPoint(
                                id=fname, image_path=path, label=label, split=split
                            )
                        )

    # add text docs
    labels = set(labels)
    for label in labels:
        data.append(
            _DataPoint(id=label, text=label.lower(), content_type='text', label=label)
        )

    # build docs
    with mp.Pool(processes=num_workers) as pool:
        docs = list(tqdm(pool.imap(_build_doc, data)))

    return DocumentArray(docs)


def _build_bird_species(root: str, num_workers: int = 8) -> DocumentArray:
    """
    Build the bird species dataset.
    Download the raw dataset from
    https://www.kaggle.com/veeralakrishna/200-bird-species-with-11788-images
    :param root: the dataset root folder.
    :param num_workers: the number of parallel workers to use.
    :return: DocumentArray
    """

    extension = '.jpg'
    root = os.path.join(root, 'CUB_200_2011', 'CUB_200_2011')
    fimages = os.path.join(root, 'images.txt')
    fclasses = os.path.join(root, 'classes.txt')
    flabels = os.path.join(root, 'image_class_labels.txt')
    fsplit = os.path.join(root, 'train_test_split.txt')
    contentdir = os.path.join(root, 'images')

    # read images.txt
    image2id = {}
    with open(fimages, 'r') as f:
        for line in f.read().splitlines():
            iid, fname = line.split()
            iid = int(iid)
            image2id[fname] = iid

    # read classes.txt
    id2class = {}
    with open(fclasses, 'r') as f:
        for line in f.read().splitlines():
            iid, classname = line.split()
            iid = int(iid)
            id2class[iid] = classname

    # read image_class_labels.txt
    imageid2classid = {}
    with open(flabels, 'r') as f:
        for line in f.read().splitlines():
            iid, cid = line.split()
            iid, cid = int(iid), int(cid)
            imageid2classid[iid] = cid

    # read train_test_split.txt
    imageid2split = {}
    with open(fsplit, 'r') as f:
        for line in f.read().splitlines():
            iid, split = line.split()
            iid, split = int(iid), int(split)
            imageid2split[iid] = split

    # add image docs
    data = []
    for rootdir, _, fnames in os.walk(contentdir):
        for fname in fnames:
            if fname.endswith(extension):
                path = os.path.join(rootdir, fname)
                image = os.path.relpath(path, contentdir)
                iid = image2id[image]
                cid = imageid2classid[iid]
                label = id2class[cid]
                split = imageid2split[iid]
                split = 'train' if split else 'test'
                data.append(
                    _DataPoint(id=fname, image_path=path, label=label, split=split)
                )

    # add text docs
    labels = {label for _, label in id2class.items()}
    for label in labels:
        data.append(
            _DataPoint(
                id=label,
                text=label[4:].lower().replace('_', ' '),
                content_type='text',
                label=label,
            )
        )

    # build docs
    with mp.Pool(processes=num_workers) as pool:
        docs = list(tqdm(pool.imap(_build_doc, data)))

    return DocumentArray(docs)


def _build_best_artworks(root: str, num_workers: int = 8) -> DocumentArray:
    """
    Build the best artworks dataset.
    Download the raw dataset from
    https://www.kaggle.com/ikarus777/best-artworks-of-all-time
    :param root: the dataset root folder.
    :param num_workers: the number of parallel workers to use.
    :return: DocumentArray
    """

    extension = '.jpg'
    fartists = os.path.join(root, 'artists.csv')
    contentdir = os.path.join(root, 'images', 'images')

    # read artists.csv
    with open(fartists, 'r') as f:
        reader = csv.reader(f)
        next(reader)
        label2genre = {row[1]: row[3] for row in reader}

    # add image docs
    data = []
    for rootdir, _, fnames in os.walk(contentdir):
        label = os.path.relpath(rootdir, contentdir).replace('_', ' ')
        for fname in fnames:
            if fname.endswith(extension):
                path = os.path.join(rootdir, fname)
                data.append(_DataPoint(id=fname, image_path=path, label=label))
        if len(fnames) > 0:
            if label == 'Albrecht Dürer':
                genre = 'Northern Renaissance'
            else:
                genre = label2genre[label]
            text = genre.lower().replace(',', ' ').replace('"', '')
            data.append(
                _DataPoint(id=genre, text=text, label=label, content_type='text')
            )

    # build docs
    with mp.Pool(processes=num_workers) as pool:
        docs = list(tqdm(pool.imap(_build_doc, data)))

    return DocumentArray(docs)


def create_file_to_text_map(dict_list):
    file_to_text = {}
    for d in dict_list:
        meta = d['metadata']
        file = meta['image'].split('//')[-1]
        attributes = meta['attributes']
        values = [d['value'] for d in attributes]
        shuffle(values)
        text = ' '.join(values)
        file_to_text[file] = text.lower()
    return file_to_text


def _build_nft(root: str, num_workers: int = 8) -> DocumentArray:
    """
    Build the nft dataset.
    Download the raw dataset from
    https://github.com/skogard/apebase
    :param root: the dataset root folder.
    :param num_workers: the number of parallel workers to use.
    :return: DocumentArray
    """
    f_labels = os.path.join(root, 'db')
    contentdir = os.path.join(root, 'ipfs')

    # read artists.csv
    with open(f_labels, 'r') as f:
        lines = f.readlines()
    dict_list = [json.loads(line) for line in lines]
    file_to_text = create_file_to_text_map(dict_list)

    data = []
    for file, text in file_to_text.items():
        data.append(_DataPoint(id=file, image_path=f'{contentdir}/{file}', label=file))
        data.append(
            _DataPoint(
                id=file + '_text',
                text=file_to_text[file],
                label=file,
                content_type='text',
            )
        )

    # build docs
    with mp.Pool(processes=num_workers) as pool:
        docs = list(tqdm(pool.imap(_build_doc, data)))

    return DocumentArray(docs)


def _build_tll(root: str, num_workers: int = 8) -> DocumentArray:
    """
    Build the tll dataset.
    Download the raw dataset from
    https://sites.google.com/view/totally-looks-like-dataset
    :param root: the dataset root folder.
    :param num_workers: the number of parallel workers to use.
    :return: DocumentArray
    """

    def transform(d: Document):
        d.load_uri_to_blob()
        d.tags['content_type'] = 'image'
        return d

    da = DocumentArray.from_files(root + '/**')
    da.apply(lambda d: transform(d))
    return da


def process_dataset(
    datadir: str,
    name: str,
    project: str,
    bucket: str,
    location: str,
    sample_k: bool = True,
    k: int = 10,
) -> None:
    """
    Build, save and upload a dataset.
    """
    root = f'{datadir}/{name}'
    out = f'{name}.bin'
    out_img10 = f'{name}.img10.bin'
    out_txt10 = f'{name}.txt10.bin'

    print(f'===> {name}')
    print(f'  Building {name} from {root} ...')
    docs = globals()[f'_build_{name.replace("-", "_")}'](root)
    docs = docs.shuffle(42)
    image_docs = DocumentArray(
        [doc for doc in docs if doc.tags['content_type'] == 'image']
    )
    text_docs = DocumentArray(
        [doc for doc in docs if doc.tags['content_type'] == 'text']
    )
    print(f'  Dataset size: {len(docs)}')
    print(f'  Num image docs: {len(image_docs)}')
    print(f'  Num text docs: {len(text_docs)}')

    if sample_k:
        print(f'  Sampling {k} image and {k} text docs ...')
        image_docs = image_docs[:k]
        text_docs = text_docs[:k]

    print('  Saving datasets ...')
    docs.save_binary(out)
    print(f'  Saved dataset to {out}')
    if sample_k:
        if len(image_docs) > 0:
            image_docs.save_binary(out_img10)
            print(f'  Saved dataset to {out_img10}')
        if len(text_docs) > 0:
            text_docs.save_binary(out_txt10)
            print(f'  Saved dataset to {out_txt10}')

    print('  Uploading datasets ...')
    upload_to_gcloud_bucket(project, bucket, location, out)
    print(f'  Uploaded dataset to gs://{bucket}/{location}/{out}')
    if sample_k:
        if len(image_docs) > 0:
            upload_to_gcloud_bucket(project, bucket, location, out_img10)
            print(f'  Uploaded dataset to gs://{bucket}/{location}/{out_img10}')
        if len(text_docs) > 0:
            upload_to_gcloud_bucket(project, bucket, location, out_txt10)
            print(f'  Uploaded dataset to gs://{bucket}/{location}/{out_txt10}')


def main():
    """
    Main method.
    """
    localdir = 'data'
    project = 'jina-simpsons-florian'
    bucket = 'jina-fashion-data'
    location = 'data/one-line/datasets/jpeg'
    datasets = [
        'tll',
        'nft-monkey',
        'deepfashion',
        'nih-chest-xrays',
        'geolocation-geoguessr',
        'stanford-cars',
        'bird-species',
        'best-artworks',
    ]
    for name in datasets:
        process_dataset(localdir, name, project, bucket, location)


if __name__ == '__main__':
    main()
